#! /usr/bin/env python3
import pathlib
from typing import Dict, Any, Sequence, Tuple, Union, Optional

from ._base import Hardware

__all__ = ["DiscreteHardware"]


class DiscreteHardware(Hardware):
    _kind: str = "discrete-hardware"

    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        self._position_identifiers: Dict[str, Any] = config.get("identifiers", {})
        super().__init__(name, config, config_filepath)

    def get_position_identifiers(self):
        return self._position_identifiers

    def set_identifier(self, identifier):
        p = self._position_identifiers[identifier]
        self.set_position(p)
        return p

    def get_identifier(self):
        return self._state["position_identifier"]



import pathlib
import subprocess
import sys
import time

import numpy as np
import pytest

import yaqc

# msgpack = yaqc.msgpack


@pytest.fixture(scope="module")
def run_daemon():
    pyfile = pathlib.Path(__file__).parent / "NdarrayTestDaemon.py"
    config = pyfile.with_suffix(".toml")
    with subprocess.Popen([sys.executable, pyfile, "--config", config]) as proc:
        start = time.time()
        while True:
            print(time.time() - start)
            try:
                client = yaqc.Client(38001)
            except ConnectionRefusedError:
                if time.time() - start > 3:
                    yield None
                    proc.terminate()
                time.sleep(0.01)
            else:
                break
        print("Connection made")
        yield client
        proc.terminate()


@pytest.mark.skip
def test_subtract_int(run_daemon):
    s = run_daemon
    result = s.subtract(minuend=np.arange(8), subtrahend=-1 * np.arange(8))
    np.testing.assert_equal(result, np.arange(8) * 2)


@pytest.mark.skip
def test_sum_float(run_daemon):
    s = run_daemon
    result = s.sum(np.linspace(0, 10))
    np.testing.assert_equal(result, np.sum(np.linspace(0, 10)))



import yaqd_core
import numpy as np


class NdarrayTestDaemon(yaqd_core.Base):
    _kind = "ndarray-test"

    def subtract(self, minuend, subtrahend):
        return minuend - subtrahend

    def sum(self, arr):
        return np.sum(arr)


if __name__ == "__main__":
    NdarrayTestDaemon.main()



#! /usr/bin/env python3

__all__ = ["Base"]

import argparse
import asyncio
import functools
import inspect
import json
import logging as logging_
import pathlib
import signal
import sys
import time
from typing import Dict, List, Optional, Any, Union, Sequence, Set

import appdirs  # type: ignore
import toml

from .__version__ import __version__, __avro_version__
from ._protocol import Protocol
from . import logging
from .exceptions import InvalidRequest, MethodNotFound

logger = logging.getLogger("yaqd_core")


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class Base:
    _daemons: List["Base"] = []
    _kind: str = "base"
    _branch: Optional[str] = None

    def __init__(self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path):
        """Create a yaq daemon.

        Parameters
        ----------
        name: str
            A name for this daemon
        config: dict
            Configuration parameters
        config_filepath: str
            The path for the configuration (not used internally, availble to clients)
        """
        self.name = name
        self._config = config
        self._config_filepath = config_filepath
        self._state_filepath = (
            pathlib.Path(appdirs.user_data_dir("yaqd-state", "yaq"))
            / self._kind
            / f"{self.name}-state.toml"
        )
        self.logger = logging.getLogger(self.name)
        if "log_level" in self._config:
            self.logger.setLevel(logging.name_to_level[self._config["log_level"]])
        if self._config.get("log_to_file"):
            fh = logging_.FileHandler(
                self._state_filepath.with_name(
                    f"{self.name}-{time.strftime('%Y-%m-%dT%H:%M:%S%z')}.log"
                )
            )
            fh.setFormatter(logging.formatter)
            self.logger.addHandler(fh)
        self.logger.info(f"Config File Path = {self._config_filepath}")
        self.logger.info(f"State File Path = {self._state_filepath}")
        self.logger.info(f"TCP Port = {config['port']}")
        self._clients: List[str] = []

        self.serial = config.get("serial", None)
        self.make = config.get("make", None)
        self.model = config.get("model", None)

        self._busy_sig = asyncio.Event()
        self._not_busy_sig = asyncio.Event()

        self._loop = asyncio.get_event_loop()

        try:
            self._state_filepath.parent.mkdir(parents=True, exist_ok=True)
            with self._state_filepath.open("rt") as f:
                state = toml.load(f)
        except (toml.TomlDecodeError, FileNotFoundError):
            state = {}

        self._load_state(state)
        self._tasks = [
            self._loop.create_task(self.save_state()),
            self._loop.create_task(self.update_state()),
        ]

    @classproperty
    def _avro_protocol(cls):
        try:
            with open(pathlib.Path(inspect.getfile(cls)).parent / f"{cls._kind}.avpr") as f:
                return json.load(f)
        except FileNotFoundError:
            return {}

    @classproperty
    def _version(cls) -> str:
        import importlib

        return getattr(
            importlib.import_module(cls.__module__.split(".")[0]),
            "__version__",
            "UNKNOWN VERSION",
        )

    @classmethod
    def main(cls):
        """Run the event loop."""
        loop = asyncio.get_event_loop()
        if sys.platform.startswith("win"):
            signals = ()
        else:
            signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            loop.add_signal_handler(s, lambda s=s: asyncio.create_task(cls.shutdown_all(s, loop)))

        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--config",
            "-c",
            default=(
                pathlib.Path(appdirs.user_config_dir("yaqd", "yaq")) / cls._kind / "config.toml"
            ),
            action="store",
            help="Path to the configuration toml file.",
        )
        parser.add_argument(
            "--verbose",
            "-v",
            action="store_const",
            dest="log_level",
            const="debug",
            help="Alias for --log-level=debug",
        )
        parser.add_argument(
            "--log-level",
            "-l",
            action="store",
            dest="log_level",
            choices=[
                "debug",
                "info",
                "notice",
                "warning",
                "error",
                "critical",
                "alert",
                "emergency",
            ],
            help="Set the log level explicitly",
        )

        parser.add_argument("--version", action="store_true")
        parser.add_argument("--protocol", action="store_true")

        args = parser.parse_args()

        if args.log_level:
            logging.setLevel(logging.name_to_level[args.log_level])

        if args.version:
            print(f"module {cls.__module__} version {cls._version}")
            print(f"avro version {__avro_version__}")
            print(f"yaqd_core version {__version__}")
            print(f"Python {sys.version}")
            sys.exit(0)

        if args.protocol:
            with open(pathlib.Path(inspect.getfile(cls)).parent / f"{cls._kind}.avpr", "r") as f:
                for line in f:
                    print(line, end="")
            sys.exit(0)

        config_filepath = pathlib.Path(args.config)
        config_file = toml.load(config_filepath)

        main_task = loop.create_task(cls._main(config_filepath, config_file, args))
        try:
            loop.run_forever()
        except asyncio.exceptions.CancelledError:
            pass
        finally:
            loop.close()

    @classmethod
    async def _main(cls, config_filepath, config_file, args=None):
        """Parse command line arguments, start event loop tasks."""
        loop = asyncio.get_running_loop()
        cls.__servers = []
        for section in config_file:
            try:
                config = cls._parse_config(config_file, section, args)
            except ValueError:
                continue
            logger.debug(f"Starting {section} with {config}")
            await cls._start_daemon(section, config, config_filepath)

        while cls.__servers:
            awaiting = cls.__servers
            cls.__servers = []
            await asyncio.wait(awaiting)
            await asyncio.sleep(1)
        loop.stop()

    @classmethod
    async def _start_daemon(cls, name, config, config_filepath):
        loop = asyncio.get_running_loop()
        daemon = cls(name, config, config_filepath)
        cls._daemons.append(daemon)

        # This function is here to namespace `daemon` so it doesn't
        # get overridden for the lambda
        def server(daemon):
            return lambda: Protocol(daemon)

        ser = await loop.create_server(
            server(daemon), config.get("host", ""), config.get("port", None)
        )
        daemon._server = ser
        cls.__servers.append(ser.serve_forever())

    @classmethod
    def _parse_config(cls, config_file, section, args=None):
        if section == "shared-settings":
            raise ValueError(f"Section name '{section}' reserved")
        config = {}
        for name, type_ in cls._avro_protocol.get("config", {}).items():
            if "default" in type_:
                config[name] = type_["default"]
        config.update(config_file.get("shared-settings", {}).copy())
        config.update(config_file[section])
        if args:
            try:
                if args.log_level:
                    config.update(log_level=args.log_level)
            except AttributeError:
                pass

        if not config.get("enable", True):
            logger.info(f"Section '{section}' is disabled")
            raise ValueError(f"Section '{section}' is disabled")
        return config

    @classmethod
    async def shutdown_all(cls, sig, loop):
        """Gracefully shutdown the asyncio loop.

        Gathers all current tasks, and allows daemons to perform cleanup tasks.

        Adapted from https://www.roguelynn.com/words/asyncio-graceful-shutdowns/
        Original code is licensed under the MIT license, and sublicensed here.
        """
        logger.info(f"Received signal {sig.name}...")
        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        logger.info(f"Cancelling {len(tasks)} outstanding tasks")
        [task.cancel() for task in tasks]
        # This is done after cancelling so that shutdown tasks which require the loop
        # are not themselves cancelled.
        [d.close() for d in cls._daemons]
        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        await asyncio.gather(*tasks, return_exceptions=True)
        [d._save_state() for d in cls._daemons]
        if hasattr(signal, "SIGHUP") and sig == signal.SIGHUP:
            config_filepath = [d._config_filepath for d in cls._daemons][0]
            config_file = toml.load(config_filepath)
            await cls._main(config_filepath, config_file)
        loop.stop()

    def shutdown(self, restart=False):
        self.logger.info(f"Shutting Down {self.name}")
        self.logger.info(f"Cancelling {len(self._tasks)} outstanding tasks")
        [task.cancel() for task in self._tasks]
        self.close()
        self._server.close()
        if restart:
            config_filepath = self._config_filepath
            config_file = toml.load(config_filepath)
            try:
                config = type(self)._parse_config(config_file, self.name)
                self._loop.create_task(
                    type(self)._start_daemon(self.name, config, config_filepath)
                )
            except ValueError as e:
                self.logger.error(e.message)

    def _connection_made(self, peername: str) -> None:
        self._clients.append(peername)
        self.logger.debug(f"_connection_made {self._clients}")

    def _connection_lost(self, peername: str) -> None:
        self._clients.remove(peername)
        self.logger.debug(f"_connection_lost {self._clients}")

    def _save_state(self) -> None:
        """Write the current state to disk."""
        with open(self._state_filepath, "wt") as f:
            f.write(self.get_state())

    async def save_state(self):
        """Schedule writing the current state to disk.

        Note: Current implementation only writes while busy (and once after busy)
        """
        while True:
            while self._busy:
                self._save_state()
                await asyncio.sleep(0.1)
            self._save_state()
            await self._busy_sig.wait()

    def get_config_filepath(self) -> str:
        """Retrieve the current filepath of the configuration."""
        return str(self._config_filepath.absolute())

    def get_config(self) -> str:
        """Retrieve the current configuration, including any defaults."""
        return toml.dumps(self._config)

    def id(self) -> Dict[str, Optional[str]]:
        """Dictionary of identifying information for the daemon."""
        return {
            "name": self.name,
            "kind": self._kind,
            "make": self.make,
            "model": self.model,
            "serial": self.serial,
        }

    @property
    def _busy(self) -> bool:
        """Indicates the current 'busy' state for use in internal functions.

        Setting busy can be done with `self._busy = <True|False>`.
        Async tasks can wait for either sense using `await self._[not_]busy_sig.wait()`.
        """
        return self._busy_sig.is_set()

    @_busy.setter
    def _busy(self, value):
        if value:
            self._busy_sig.set()
            self._not_busy_sig.clear()
        else:
            self._not_busy_sig.set()
            self._busy_sig.clear()

    def busy(self) -> bool:
        """Boolean representing if the daemon is busy (state updated) or not."""
        return self._busy

    # The following functions (plus __init__) are what most daemon need to implement

    async def update_state(self):
        """Continually monitor and update the current daemon state."""
        pass

    def get_state(self) -> str:
        """Return the current daemon state."""
        return toml.dumps(self._state)

    def _load_state(self, state):
        """Load an initial state from a dictionary (typically read from the state.toml file).

        Must be tolerant of missing fields, including entirely empty initial states.
        Raise an exception if state is invalid.

        Parameters
        ----------
        state: dict
            The saved state to load.
        """
        self._state = state
        for name, type_ in self._avro_protocol.get("state", {}).items():
            self._state.setdefault(name, type_.get("default", None))

    def close(self):
        pass



import pathlib
from typing import Dict, Any, Sequence, List, Union, Optional

from ._base import Hardware

__all__ = ["ContinuousHardware"]


class ContinuousHardware(Hardware):
    _kind: str = "continuous-hardware"

    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self._out_of_limits = config["out_of_limits"]

    def get_limits(self) -> List[float]:
        assert self._state["hw_limits"][0] < self._state["hw_limits"][1]
        config_limits = self._config["limits"]
        assert config_limits[0] < config_limits[1]
        out = [
            max(self._state["hw_limits"][0], config_limits[0]),
            min(self._state["hw_limits"][1], config_limits[1]),
        ]
        assert out[0] < out[1]
        return out

    def in_limits(self, position: float) -> bool:
        low, upp = self.get_limits()
        if low <= position <= upp:
            return True
        else:
            return False

    def set_position(self, position: float) -> None:
        if not self.in_limits(position):
            if self._out_of_limits == "closest":
                low, upp = self.get_limits()
                if position > upp:
                    position = upp
                elif position < low:
                    position = low
            elif self._out_of_limits == "ignore":
                return
            else:
                raise ValueError(f"{position} not in ranges {self.get_limits()}")
        super().set_position(position)



__all__ = ["ASerial"]


import asyncio
import serial  # type: ignore


class ASerial(serial.Serial):
    def __init__(self, port=None, baudrate=9600, *args, **kwargs):
        super().__init__(port, baudrate, timeout=0, *args, **kwargs)
        self._readlock = asyncio.Lock()

    async def aread(self, size=1):
        async with self._readlock:
            return await self._aread(size)

    async def _aread(self, size=1):
        buf = b""
        while len(buf) < size:
            buf += self.read(size - len(buf))
            await asyncio.sleep(0.01)
        return buf

    async def areadline(self, size=-1):
        async with self._readlock:
            return await self._areadline(size)

    async def _areadline(self, size=-1):
        buf = b""
        while not buf.endswith(b"\n") and (size < 0 or len(buf) < size):
            buf += self.readline(size - len(buf) if size >= 0 else -1)
            await asyncio.sleep(0.01)
        return buf

    async def areadlines(self, hint=-1):
        async with self._readlock:
            size = 0
            while hint < 0 or size < hint:
                s = await self._areadline()
                size += len(s)
                yield s

    async def awrite_then_read(self, data, size=1):
        async with self._readlock:
            self.write(data)
            return await self._aread(size)

    async def awrite_then_readline(self, data, size=-1):
        async with self._readlock:
            self.write(data)
            return await self._areadline(size)



import pathlib
import subprocess
import sys

import pytest

here = pathlib.Path(__file__).parent


def test_malformed():
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(
            [sys.executable, here / "config.py", "--config", here / "malformed.toml",],
            check=True,
            timeout=0.5,
        )



"""Core python package for implementing yaq deamons, and associated utilities."""

__all__ = [
    "exceptions",
    "logging",
    "__version__",
    "Base",
    "Hardware",
    "ContinuousHardware",
    "Sensor",
    "DiscreteHardware",
]

from . import exceptions
from . import logging
from .__version__ import __version__
from ._daemon import Base
from ._hardware import Hardware, ContinuousHardware, DiscreteHardware
from ._sensor import Sensor



__all__ = ["Sensor"]


import asyncio
import pathlib
from typing import Dict, Any, Sequence, Union, Tuple, List, Optional
from ._daemon import Base

# TODO: add array type
MeasureType = Dict[str, Union[float]]


class Sensor(Base):
    _kind = "base-sensor"

    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self._measured: MeasureType = dict()  # values must be numbers or arrays
        self._channel_names: List[str] = []
        self._channel_units: Dict[str, str] = dict()
        self._channel_shapes: Dict[str, Tuple[int]] = dict()
        self._measurement_id = 0

    def measure(self, loop: bool = False) -> int:
        """Start a measurement, optionally looping.

        Sensor will remain busy until measurement completes.

        Parameters
        ----------
        loop: bool, optional
            Toggle looping behavior. Default False.

        See Also
        --------
        stop_looping
        """
        self._looping = loop
        if not self._busy:
            self._busy = True
            self._loop.create_task(self._runner(loop=loop))
        return self._measurement_id

    def get_channel_names(self):
        """Get current channel names."""
        return self._channel_names

    def get_channel_shapes(self):
        """Get channel shapes."""
        # as default behavior, assume all channels are scalars
        if self._channel_shapes:
            return self._channel_shapes
        else:
            return {k: () for k in self._channel_names}

    def get_channel_units(self):
        """Get channel units."""
        return self._channel_units

    def get_measured(self) -> MeasureType:
        """Get most recently measured values."""
        return self._measured

    async def _measure(self) -> MeasureType:
        """Do measurement, filling _measured dictionary.

        Returns dictionary with keys channel names, values numbers or arrays.
        """
        raise NotImplementedError

    async def _runner(self, loop: bool) -> None:
        """Handle execution of _measure, including looping and setting of _measurement_id."""
        while True:
            self._measured = await self._measure()
            assert set(self._measured.keys()) == set(self._channel_names)
            self._measured["measurement_id"] = self._measurement_id
            if not self._looping:
                self._busy = False
                self._measurement_id += 1
                break
            await asyncio.sleep(0)

    def stop_looping(self) -> None:
        """Stop looping."""
        self._looping = False



import fastavro  # type: ignore


class Interface:
    def __init__(self, array_interface):
        array_interface["shape"] = tuple(array_interface["shape"])
        self.__array_interface__ = array_interface


def read_ndarray(data, writer_schema, reader_schema):
    import numpy as np  # type: ignore

    return np.array(Interface(data))


def prepare_ndarray(data, schema):
    if hasattr(data, "__array_interface__"):
        array_interface = data.__array_interface__.copy()
        array_interface["data"] = data.tobytes()
        array_interface["shape"] = list(array_interface["shape"])
        return array_interface
    else:
        return data


fastavro.read.LOGICAL_READERS["record-ndarray"] = read_ndarray
fastavro.write.LOGICAL_WRITERS["record-ndarray"] = prepare_ndarray



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['yaqd_core',
 'yaqd_core._hardware',
 'yaqd_core.aserial',
 'yaqd_core.avrorpc',
 'yaqd_core.exceptions',
 'yaqd_core.logging']

package_data = \
{'': ['*']}

install_requires = \
['appdirs', 'toml', 'fastavro']

extras_require = \
{'dev': ['black', 'pre-commit']}

setup(name='yaqd-core',
      version='2020.7.1',
      description='Core python package for implementing yaq deamons, and associated utilities.',
      author='yaq developers',
      author_email='git@ksunden.space',
      url='https://yaq.fyi',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



__all__ = ["Unpacker"]

import asyncio
from collections import namedtuple
import io
import struct
import warnings

import fastavro  # type: ignore

from .handshake import handle_handshake, handshake_request_schema
from .. import logging


class Unpacker:
    def __init__(self, protocol, file_like=None):
        self.protocol = protocol
        if file_like is None:
            self._file = io.BytesIO()
        else:
            self._file = file_like
        self.buf = io.BytesIO()
        self.handshake_complete = False
        self.handshake_response = None
        self.meta = None
        self.message_name = None
        self.parameters = None
        self.remaining = 0

    def __iter__(self):
        return self

    def __next__(self):
        try:
            if not self.handshake_complete and self.handshake_response is None:
                handshake_request = self._read_object(handshake_request_schema)
                self.handshake_response = handle_handshake(handshake_request, self.protocol)
                if self.handshake_response.match == "BOTH":
                    self.handshake_complete = True
            if self.meta is None:
                self.meta = self._read_object({"type": "map", "values": "bytes"})
            if self.message_name is None:
                self.message_name = self._read_object("string")
            if self.message_name != "" and self.protocol["messages"][self.message_name].get(
                "request", []
            ):
                self._read_parameters(self.message_name)

            ret = (
                self.handshake_response,
                self.meta,
                self.message_name,
                self.parameters,
            )
            self.handshake_response = None
            self.meta = None
            self.message_name = None
            self.parameters = None
            return ret

        except (ValueError, struct.error):
            raise StopIteration

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                return next(self)
            except StopIteration:
                await asyncio.sleep(0.001)

    def feed(self, data: bytes):
        # Must support random access, if it does not, must be fed externally (e.g. TCP)
        pos = self._file.tell()
        self._file.seek(0, 2)
        self._file.write(data)
        self._file.seek(pos)

    def _read_object(self, schema):
        while True:
            try:
                self.buf.seek(0)
                obj = fastavro.schemaless_reader(self.buf, schema)
                self.buf = io.BytesIO()
                return obj
            except Exception as e:
                self.buf.seek(0)
                pass
            if not self.remaining:
                self.remaining = struct.unpack_from(">L", self._file.read(4))[0]

            self.buf.seek(0, 2)
            num_read = self.buf.write(self._file.read(self.remaining))
            self.remaining -= num_read

    def _read_parameters(self, name):
        if self.parameters is None:
            self.parameters = []
        for param_schema in self.protocol["messages"][name]["request"][len(self.parameters) :]:
            self.parameters.append(self._read_object(param_schema["type"]))



__all__ = ["Hardware"]

import math
import pathlib
from typing import Dict, Any, Union, Optional

from .._daemon import Base


class Hardware(Base):
    traits = ["has-position"]
    _kind = "hardware"

    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        self._units = None
        super().__init__(name, config, config_filepath)

    def get_position(self) -> float:
        return self._state["position"]

    def get_units(self) -> Union[str, None]:
        return self._units

    def get_destination(self) -> float:
        return self._state["destination"]

    def set_position(self, position: float) -> None:
        self._busy = True
        self._state["destination"] = position
        self._set_position(position)

    def _set_position(self, position: float) -> None:
        raise NotImplementedError

    def set_relative(self, distance: float) -> float:
        new = self._state["destination"] + distance
        self.set_position(new)
        return new



import pathlib
import subprocess
import sys
import time

import pytest

import yaqc
import yaqd_core


@pytest.fixture(scope="module")
def run_daemon():
    config = pathlib.Path(__file__).parent / "shutdown.toml"
    pyfile = config.with_suffix(".py")
    with subprocess.Popen([sys.executable, pyfile, "--config", config]) as proc:
        while True:
            try:
                clients = yaqc.Client(39098), yaqc.Client(39099)
            except ConnectionRefusedError:
                time.sleep(0.01)
            else:
                break
        yield clients

        proc.terminate()


def test_shutdown(run_daemon):
    restart, shutdown = run_daemon
    assert shutdown.id()["name"] == "shutdown"
    assert restart.id()["name"] == "restart"

    shutdown.shutdown()
    with pytest.raises(ConnectionError):
        shutdown.id()


def test_restart(run_daemon):
    restart, shutdown = run_daemon
    assert restart.id()["name"] == "restart"

    restart.shutdown(restart=True)
    time.sleep(0.1)
    start = time.time()
    while time.time() - start < 3:
        try:
            restart = yaqc.Client(39098)
            assert restart.id()["name"] == "restart"
        except (ConnectionError):
            time.sleep(0.1)
        else:
            break
    else:
        raise TimeoutError



__all__ = [
    "DEBUG",
    "INFO",
    "NOTICE",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "ALERT",
    "EMERGENCY",
    "getLogger",
    "name_to_level",
    "setLevel",
]

import logging


# Log levels as defined by sd-daemon(3)
DEBUG = logging.DEBUG
INFO = logging.INFO
NOTICE = logging.INFO + 5
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL
ALERT = logging.CRITICAL + 10
EMERGENCY = logging.CRITICAL + 20

name_to_level = {
    "debug": DEBUG,
    "info": INFO,
    "notice": NOTICE,
    "warning": WARNING,
    "error": ERROR,
    "critical": INFO,
    "alert": ALERT,
    "emergency": EMERGENCY,
}

# Rename levels to start with numeric prefix following syslog convention
logging.addLevelName(DEBUG, "<7> DEBUG")
logging.addLevelName(INFO, "<6> INFO")
logging.addLevelName(NOTICE, "<5> NOTICE")
logging.addLevelName(WARNING, "<4> WARNING")
logging.addLevelName(ERROR, "<3> ERR")
logging.addLevelName(CRITICAL, "<2> CRIT")
logging.addLevelName(ALERT, "<1> ALERT")
logging.addLevelName(EMERGENCY, "<0> EMERG")

formatter = logging.Formatter(
    "{levelname} : {asctime} : {name} : {message}",
    datefmt="%Y-%m-%dT%H:%M:%S%z",
    style="{",
)

_loggers = []
_default_level = logging.INFO


def getLogger(name=None):
    """Wrapper of `logging.getLogger` which sets the default formatter."""
    logger = logging.getLogger(name)
    if not logger.hasHandlers():
        ch = logging.StreamHandler()
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        logger.setLevel(_default_level)
        _loggers.append(logger)
    return logger


def setLevel(level):
    global _default_level
    _default_level = level
    for log in _loggers:
        log.setLevel(level)



import asyncio
import io
import struct

import fastavro  # type: ignore
from . import avrorpc

from .exceptions import InvalidRequest, MethodNotFound


class Protocol(asyncio.Protocol):
    def __init__(self, daemon, *args, **kwargs):
        self._daemon = daemon
        self.logger = daemon.logger
        self._avro_protocol = daemon._avro_protocol
        asyncio.Protocol.__init__(self, *args, **kwargs)

    def connection_lost(self, exc):
        peername = self.transport.get_extra_info("peername")
        self.logger.info(f"Connection lost from {peername} to {self._daemon.name}")
        self._daemon._connection_lost(peername)

    def connection_made(self, transport):
        """Process an incomming connection."""
        peername = transport.get_extra_info("peername")
        self.logger.info(f"Connection made from {peername} to {self._daemon.name}")
        self.transport = transport
        self.unpacker = avrorpc.Unpacker(self._avro_protocol)
        self._daemon._connection_made(peername)

    def data_received(self, data):
        """Process an incomming request."""
        self.logger.debug(f"Data received: {repr(data)}")
        if not self._daemon._server.is_serving():
            self.transport.close()
        self.unpacker.feed(data)
        try:
            for hs, meta, name, params in self.unpacker:
                if hs is not None:
                    out = bytes(hs)
                    out = struct.pack(">L", len(out)) + out
                    self.transport.write(out)
                    if hs.match == "NONE":
                        name = ""

                out_meta = io.BytesIO()
                fastavro.schemaless_writer(out_meta, {"type": "map", "values": "bytes"}, meta)
                length = out_meta.tell()
                self.transport.write(struct.pack(">L", length) + out_meta.getvalue())
                self.logger.debug(f"Wrote meta, {meta}, {out_meta.getvalue()}")
                try:
                    response_out = io.BytesIO()
                    response = None
                    response_schema = "null"
                    if name:
                        fun = getattr(self._daemon, name)
                        if params is None:
                            params = []
                        response = fun(*params)
                        response_schema = self._avro_protocol["messages"][name].get(
                            "response", "null"
                        )
                    fastavro.schemaless_writer(response_out, response_schema, response)
                except Exception as e:
                    self.logger.error(f"Caught exception: {type(e)} in message {name}")
                    self.transport.write(struct.pack(">L", 1) + b"\1")
                    error_out = io.BytesIO()
                    fastavro.schemaless_writer(error_out, ["string"], repr(e))
                    length = error_out.tell()
                    self.transport.write(struct.pack(">L", length) + error_out.getvalue())
                else:
                    self.transport.write(struct.pack(">L", 1) + b"\0")
                    self.logger.debug(f"Wrote non-error flag")
                    length = response_out.tell()
                    self.transport.write(struct.pack(">L", length) + response_out.getvalue())
                    self.logger.debug(f"Wrote response {response}, {response_out.getvalue()}")
                self.transport.write(struct.pack(">L", 0))
                if name == "shutdown":
                    self.logger.debug("Closing transport")
                    self.transport.close()

        except (ValueError):
            pass
